<?
/* QM Promisance - Turn-based strategy game
 * Copyright (C) 2001-2009 QMT Productions
 *
 * $URL: https://qmpromisance.svn.sourceforge.net/svnroot/qmpromisance/promisance/tags/release-4_3/classes/prom_entity.php $
 * $Id: prom_entity.php 1692 2009-11-01 20:48:01Z Quietust $
 */

if (!defined('IN_GAME'))
	die("Access denied");

require_once(PROM_BASEDIR .'includes/database.php');

abstract class prom_entity implements IteratorAggregate
{
	protected $db_table;	// Database table name
	protected $db_id;	// Database table primary key column
	protected $db_entid;	// Database lock table entity type ID

	protected $id;		// entity ID for internal tracking

	protected $data;	// __get and __set will automatically look into this - use getData/setData otherwise

	protected $lockid;	// indicates who has locked the record
	protected $needlock;	// used exclusively by the turns script, allows modification without locking

	protected $update;	// list of fields within $data that have been modified

	// Constructor - initializes entity with a specific ID
	protected function __construct($id, $needlock, $db_table, $db_id, $db_entid)
	{
		$this->id = $id;
		$this->needlock = $needlock;
		$this->db_table = $db_table;
		$this->db_id = $db_id;
		$this->db_entid = $db_entid;

		$this->data = array();
		$this->update = array();
		$this->lockid = 0;
	}

	// Load data into entity
	public function initdata ($data)
	{
		if (!$data)
			return FALSE;

		$this->data = $data;
		$this->update = array();
		return TRUE;
	}

	// Fetch entity data from the database
	public function load ()
	{
		if ($this->id == 0)
		{
			warning('Attempted to load uninitialized entity of type '. get_class($this), 1);
			return FALSE;
		}

		global $db;

		$q = $db->prepare('SELECT * FROM '. $this->db_table .' WHERE '. $this->db_id .' = ?');
		$q->bindIntValue(1, $this->id);
		if (!$q->execute())
		{
			warning('Entity '. $this->id .' of type '. get_class($this) .' failed to load from database', 1);
			return FALSE;
		}

		if (!$this->initdata($q->fetch()))
			return FALSE;

		$this->update = array();
		$this->lockid = 0;
		return TRUE;
	}

	// Query whether or not the current record is locked
	public function locked ()
	{
		return ($this->lockid != 0);
	}

	// Attempt to lock the entity within the database
	public function lock ($owner)
	{
		if ($this->id == 0)
		{
			warning('Attempted to lock uninitialized entity of type '. get_class($this), 1);
			return FALSE;
		}

		if ($this->locked())
		{
			// not strictly a failure, since it's already locked, but should still generate a warning message
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to lock while already locked', 1);
			return TRUE;
		}

		if ($this->needlock)
		{
			global $db;

			$q = $db->prepare('INSERT INTO '. LOCK_TABLE .' (lock_id,lock_type,lock_owner) VALUES (?,?,?)');
			$q->bindIntValue(1, $this->id);
			$q->bindIntValue(2, $this->db_entid);
			$q->bindIntValue(3, $owner);
			
			// failure to insert likely means there was a key violation, meaning it was already locked
			if (!$q->execute())
				return FALSE;
		}
		// if the entity locked properly, then load it and set lock owner
		if (!$this->load())
		{
			// load failed, but lock was successful, so we NEED to clear it
			// set lock ID so unlock doesn't bail out immediately
			$this->lockid = $owner;
			$this->unlock();
			return FALSE;
		}
		$this->lockid = $owner;

		return TRUE;
	}

	// Release the lock on an entity within the database
	public function unlock ()
	{
		if ($this->id == 0)
		{
			warning('Attempted to unlock uninitialized entity of type '. get_class($this), 1);
			return FALSE;
		}

		if (!$this->locked())
		{
			// not strictly a failure, since it's already locked, but should still generate a warning message
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to unlock while not locked', 1);
			return TRUE;
		}

		if ($this->needlock)
		{
			global $db;

			$q = $db->prepare('DELETE FROM '. LOCK_TABLE .' WHERE lock_id = ? AND lock_type = ? AND lock_owner = ?');
			$q->bindIntValue(1, $this->id);
			$q->bindIntValue(2, $this->db_entid);
			$q->bindIntValue(3, $this->lockid);
			if (!$q->execute())
			{
				warning('Entity '. $this->id .' of type '. get_class($this) .' failed to unlock', 1);
				return FALSE;
			}

			// can happen if a lock is forcibly released from admin/locks.php
			// this should still return TRUE and mark the entity as unlocked
			if ($q->rowCount() == 0)
				warning('Entity '. $this->id .' of type '. get_class($this) .' abruptly lost database lock', 1);
		}
		$this->lockid = 0;
		return TRUE;
	}

	// Magic function to allow retrieving data values directly
	public function __get ($field)
	{
		return $this->_getData($field);
	}

	// Magic function to allow modifying data values directly
	// and automatically including them in the list of fields to synch to the database
	public function __set ($field, $value)
	{
		$this->_setData($field, $value);
	}

	// Magic function to allow checking for presence of data fields
	public function __isset ($field)
	{
		return isset($this->data[$field]);
	}

	// Under no circumstances should a data field ever be unset
	// Display a warning if anything tries to do so
	public function __unset ($field)
	{
		warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to unset data field "'. $field .'"', 1);
	}

	// Legacy function for getting data fields
	public function getData ($field)
	{
		return $this->_getData($field);
	}

	// Legacy function for setting data fields
	public function setData ($field, $value)
	{
		return $this->_setData($field, $value);
	}

	// Underlying function for getting data fields
	protected function _getData ($field)
	{
		if ($this->id == 0)
		{
			warning('Attempted to read from uninitialized entity of type '. get_class($this), 2);
			return FALSE;
		}

		if (!isset($this->data[$field]))
		{
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to retrieve nonexistant data field "'. $field .'"', 2);
			return NULL;
		}

		return $this->data[$field];
	}

	// Underlying function for setting data fields
	protected function _setData ($field, $value)
	{
		if ($this->id == 0)
		{
			warning('Attempted to write to uninitialized entity of type '. get_class($this), 2);
			return FALSE;
		}

		if (!$this->locked())
		{
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to modify data field "'. $field .'" while not locked', 2);
			return FALSE;
		}

		if (!isset($this->data[$field]))
		{
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to modify nonexistant data field "'. $field .'"', 2);
			return FALSE;
		}

		$this->update[] = $field;
		$this->data[$field] = $value;
		return TRUE;
	}

	// Saves an entity's data back to the database
	// Specify FALSE to keep it locked afterwards
	public function save ($unlock = TRUE)
	{
		if ($this->id == 0)
		{
			warning('Attempted to save uninitialized entity of type '. get_class($this), 1);
			return FALSE;
		}
		if (!$this->locked())
		{
			warning('Entity '. $this->id .' of type '. get_class($this) .' attempted to save while not locked', 1);
			return FALSE;
		}
		// if there are no fields to update, simply perform an unlock and return
		if (count($this->update) == 0)
			return $unlock ? $this->unlock() : TRUE;

		$this->update = array_unique($this->update);

		$update_fields = array();
		$update_values = array();
		foreach ($this->update as $field)
		{
			$update_fields[] = $field .'=?';
			$update_values[] = $this->data[$field];
		}
		$update_query = implode(',', $update_fields);

		global $db;

		$q = $db->prepare('UPDATE '. $this->db_table .' SET '. $update_query .' WHERE '. $this->db_id .' = ?');
		$update_values[] = $this->id;
		$q->bindAllValues($update_values);
		if (!$q->execute())
		{
			$this->update[] = $this->db_id;		// need this for proper error reporting
			warning('Entity '. $this->id .' of type '. get_class($this) .' failed to save to database', 1, var_export(array_combine($this->update, $update_values), true));
			if ($unlock)
				$this->unlock();
			return FALSE;
		}
		$this->update = array();
		return $unlock ? $this->unlock() : TRUE;
	}

	// Allows using foreach() to iterate across properties
	public function getIterator()
	{
		return new ArrayIterator($this->data);
	}

	// Attempt to lock all of the specified entities
	public static function lock_entities ($ents, $owner)
	{
		global $db;
		// to start off, make sure everything is actually a lockable entity
		foreach ($ents as $ent)
		{
			if (!($ent instanceof prom_entity))
				error_500('ERROR_TITLE', 'Attempted to request lock on a non-entity!');
			if ($ent->db_entid == 0)
				error_500('ERROR_TITLE', 'Attempted to request lock on a non-lockable entity!');
		}
		$start = microtime(TRUE);
		while (1)
		{
			// lock entities one by one
			$needlock = count($ents);
			foreach ($ents as $ent)
			{
				if (!$ent->lock($owner))
					break;
				$needlock--;
			}
			if ($needlock == 0)
				break;
			// if one lock fails, relinquish all of the ones that succeeded
			foreach ($ents as $ent)
				if ($ent->locked())
					$ent->unlock();

			// delay 0.1 to 0.2 seconds, then try again
			usleep(mt_rand(100000, 200000));
			// and give up entirely after 30 seconds - by then, something has gone wrong
			if (microtime(TRUE) > $start + 30)
				error_503('ERROR_TITLE', lang('ERROR_LOCK_TIMEOUT'));
		}
	}

	// Forcibly lock all entities of a particular type
	protected static function lock_all_entities ($table, $col, $type, $owner)
	{
		global $db;
		while (1)
		{
			// MySQL supports the syntax "INSERT IGNORE", which we can use to lock all unlocked entities in bulk
			if ($db->getAttribute(PDO::ATTR_DRIVER_NAME) == 'mysql')
			{
				$q = $db->prepare('INSERT IGNORE INTO '. LOCK_TABLE .' (lock_id,lock_type,lock_owner) SELECT '. $col .',?,? FROM '. $table);
				$q->bindIntValue(1, $type);
				$q->bindIntValue(2, $owner);
				$q->execute() or warning('Failed to mass lock entities of type '. $table, 2);
			}
			else	// However, other databases do not, so for those we need to add the locks one at a time
			{
				$q = $db->prepare('SELECT '. $col .' FROM '. $table .' WHERE '. $col .' NOT IN (SELECT lock_id FROM '. LOCK_TABLE .' WHERE lock_type = ?)');
				$q->bindIntValue(1, $type);
				$q->execute() or warning('Failed to acquire lock list for entities of type '. $table, 2);
				$ids = $q->fetchAll();
				$q = $db->prepare('INSERT INTO '. LOCK_TABLE .' (lock_id,lock_type,lock_owner) VALUES (?,?,?)');
				$q->bindIntValue(2, $type);
				$q->bindIntValue(3, $owner);
				foreach ($ids as $id)
				{
					$q->bindIntValue(1, $id[$col]);
					$q->execute();
				}
			}
			if ($db->queryCell('SELECT COUNT(*) FROM '. $table .' WHERE '. $col .' NOT IN (SELECT lock_id FROM '. LOCK_TABLE .' WHERE lock_type = ? AND lock_owner = ?)', array($type, $owner)) == 0)
				break;
			// delay 50-100ms and retry if any have not been locked by specified owner
			usleep(mt_rand(50000, 100000));
		}
	}
} // abstract class prom_entity
?>
