<?
/* QM Promisance - Turn-based strategy game
 * Copyright (C) 2001-2009 QMT Productions
 *
 * $URL: https://qmpromisance.svn.sourceforge.net/svnroot/qmpromisance/promisance/tags/release-4_3/includes/database.php $
 * $Id: database.php 1692 2009-11-01 20:48:01Z Quietust $
 */

if (!defined('IN_GAME'))
	die("Access denied");

// Supported database types:
// 'mysql' - MySQL 4.1 or greater
// 'pgsql' - PostgreSQL 8.1 or greater

function db_open ($type, $sock, $host, $port, $user, $pass, $name)
{
	if (!strlen($sock) && !strlen($host))
		die('Invalid database configuration - must specify either hostname or UNIX socket');
	if (strlen($sock) && strlen($host))
		die('Invalid database configuration - cannot specify both hostname and UNIX socket');

	// Select database driver and construct DSN for connection
	switch ($type)
	{
	case 'mysql':
		$dsn = 'mysql:';
		if ($sock)
			$dsn .= 'unix_socket='. $sock .';';
		elseif ($host)
		{
			$dsn .= 'host='. $host .';';
			if ($port)
				$dsn .= 'port='. $port .';';
		}
		$dsn .= 'dbname='. $name;
		break;
	case 'pgsql':
		$dsn = 'pgsql:';
		if ($sock)
			$dsn .= 'host='. $sock .' ';
		elseif ($host)
		{
			$dsn .= 'host='. $host .' ';
			if ($port)
				$dsn .= 'port='. $port .' ';
		}
		$dsn .= 'dbname='. $name;
		break;
/*	// The SQLite PDO driver in PHP 5.2 is broken - see http://bonsai.php.net/bug.php?id=45259 for details
	// The exact effects are not known, though previous functionality had been broken by this
	case 'sqlite':
		if ($host)
			die('SQLite driver must specify filename in SOCK configuration value');
		$dsn = 'sqlite:'. $sock;
		break;
*/
	default:
		die('An unsupported database driver has been specified!');
	}

	// Attempt to open database connection
	try
	{
		$db = new QM_PDO($dsn, $user, $pass);
	}
	catch (PDOException $e)
	{
		return NULL;
	}

	return $db;
}

// Wrapper for PDO, adds query counting and some additional methods for common database-bound actions
class QM_PDO extends PDO
{
	public $queryCount;	// needs to be public so QM_PDOStatement can update it easily

	public function __construct($dsn, $username = '', $password = '', $driver_options = array())
	{
		parent::__construct($dsn, $username, $password, $driver_options);
		$this->setAttribute(PDO::ATTR_STATEMENT_CLASS, array('QM_PDOStatement', array($this)));
		$this->setAttribute(PDO::ATTR_ERRMODE, PDO::ERRMODE_SILENT);

		// Database-specific post-connect actions
		switch ($this->getAttribute(PDO::ATTR_DRIVER_NAME))
		{
		case 'mysql':
			$this->exec("SET NAMES 'utf8'");
			$this->exec("SET SESSION sql_mode=''");
			break;
		case 'sqlite':
			$this->exec("PRAGMA encoding = 'UTF-8'");
			$this->exec("PRAGMA locking_mode = EXCLUSIVE");
			$this->exec("PRAGMA synchronous = OFF");
			$this->sqliteCreateFunction('LEAST', array('QM_PDO', 'sqlite_least'), 2);
			break;
		}
		$this->queryCount = 0;
	}

	// Placeholder for query logging
	public function query ($statement)
	{
		$this->queryCount++;
		return parent::query($statement);
	}

	// 1-liner for executing quick SQL queries, intended only for turns.php
	public function queryParam ($query, $parms = NULL)
	{
		$q = $this->prepare($query) or warning('SQL prepare failed', 1);
		if ($parms)
			$q->bindAllValues($parms);
		$q->execute() or warning('SQL execute failed', 1);
		return $q;
	}

	// evaluate an SQL query, return first cell of first row
	// useful for "SELECT COUNT(*) ..." queries
	public function queryCell ($query, $parms = NULL)
	{
		$q = $this->queryParam($query, $parms);
		return $q->fetchColumn();
	}

	// Clear the contents of a table and reset its auto-incrementing ID
	public function clearTable ($table)
	{
		switch ($this->getAttribute(PDO::ATTR_DRIVER_NAME))
		{
		case 'mysql':
			if (!$this->query('TRUNCATE TABLE '. $table))
				return FALSE;
			break;
		case 'pgsql':
			if (!$this->query('TRUNCATE TABLE '. $table))
				return FALSE;
			$seq = $this->getSequence($table);
			if ($seq)
			{
				if (!$this->query('ALTER SEQUENCE '. $seq .' RESTART WITH 1'))
					return FALSE;
			}
			break;
		case 'sqlite':
			if (!$this->query('DELETE FROM '. $table))
				return FALSE;
			break;
		}
		return TRUE;
	}

	// Retrieves a database's sequence, if it exists
	public function getSequence ($table)
	{
		global $sequences;
		// MySQL and SQLite do not use sequences
		if (in_array($this->getAttribute(PDO::ATTR_DRIVER_NAME), array('mysql', 'sqlite')))
			return NULL;
		// Not all tables have sequences associated with them
		if (!isset($sequences[$table]))
			return NULL;
		return $sequences[$table];
	}

	// Modifies the query to limit rows returned
	public function setLimit ($query, $rows, $offset = 0)
	{
		switch ($this->getAttribute(PDO::ATTR_DRIVER_NAME))
		{
		case 'mysql':
			if ($offset)
				$query .= ' LIMIT '. $offset .','. $rows;
			else	$query .= ' LIMIT '. $rows;
			break;
		case 'pgsql':
		case 'sqlite':
			$query .= ' LIMIT '. $rows;
			if ($offset)
				$query .= ' OFFSET '. $offset;
			break;
		}
		return $query;
	}

	// SQLite does not implement the LEAST() function (it uses MIN() instead), so it needs to be implemented here
	public static function sqlite_least ($num1, $num2)
	{
		return min($num1, $num2);
	}

}

// Wrapper for PDOStatement, adds query counting and some shortcuts for binding parameters
class QM_PDOStatement extends PDOStatement
{
	protected $dbh;
	protected function __construct ($dbh)
	{
		$this->dbh = $dbh;
	}

	// Placeholder for query logging
	public function execute ($input_parameters = NULL)
	{
		$this->dbh->queryCount++;
		return parent::execute($input_parameters);
	}

	// Shortcut for binding an integer parameter
	public function bindIntValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_INT);
	}

	// Shortcut for binding a string parameter
	public function bindStrValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_STR);
	}

	// Shortcut for binding a float parameter
	// Since PDO lacks a "float" type, bind as string for now
	public function bindFltValue ($parm, $value)
	{
		return $this->bindValue($parm, $value, PDO::PARAM_STR);
	}

	// Automatically bind all parameters based on type (for cases where the parameter count varies)
	public function bindAllValues ($parms)
	{
		foreach ($parms as $i => $val)
		{
			if (is_numeric($val))
			{
				if ($val - floor($val) > 0.00001)
					$this->bindFltValue($i + 1, $val);
				else	$this->bindIntValue($i + 1, $val);
			}
			else	$this->bindStrValue($i + 1, $val);
		}
	}
}
?>
